import urllib2
from logging import getLogger
from rdflib.Graph import ConjunctiveGraph
from rdflib import URIRef, Literal
from rdflib.store.SPARQL import SPARQLStore
from curate.work import queue

def httpGET(_unusedp, resourcep):
    """
    :param resourcep: resource pattern, the bound version of which
        checked

    Check if the given resource responds with data to an HTTP
    GET request.
    """
    log = getLogger("httpGET(%s, %s)" % (_unusedp.n3(), resourcep.n3()))
    log.debug("init")
    def f(_unused, resource):
        try:
            log.info(resource)
            fp = urllib2.urlopen(resource)
            fp.read()
            fp.close()
            return True
        except Exception, e:
            log.warn("%s", e)
            return False
    return f

def sparqlCheck(_unusedp, endpointp):
    """
    :param endpointp: endpoint pattern, the bound version of which
        will be checked

    Check if the SPARQL endpoint functions. In this context, to
    *function* means returning a result when given a simple query.
    The query used is:

        SELECT * WHERE { ?s ?p ?o } LIMIT 1
    """
    log = getLogger("sparqlCheck(%s, %s)" % (_unusedp.n3(), endpointp.n3()))
    log.debug("init")
    def f(_unused, endpoint):
        store = SPARQLStore(endpoint)
        log.info(endpoint)
        try:
            ## as generic a query as possible
            q = "SELECT * WHERE { ?s ?p ?o } LIMIT 1"
            for x in ConjunctiveGraph(store).query(q):
                return True
        except Exception, e:
            log.warn("%s", e)
        return False
    return f

def addTag(datasetp, tagp):
    """
    :param datasetp: the dataset variable, the bound version of which
        will acquire the tag
    :param tagp: the tag variable, the bound version of which will be
        added to the dataset

    This builtin adds a tag to a dataset via the CKAN api. The tag is
    expected to be a literal value.
    """
    def f(dataset, tag):
        queue.add(dataset, "tags", [unicode(tag)])
        return True
    return f

def addGroup(datasetp, groupp):
    """
    :param datasetp: the dataset variable, the bound version of which
        will be added to the specified group
    :param groupp: the group variable, the bound version of which will
        have the dataset added to.

    This builtin adds a dataset to a group. The group can be a  URI or
    a literal. If it is a URI, it will be split on the / character and
    the dataset will be added to the group named with the last component
    of the path.
    """
    def f(dataset, group):
        if isinstance(group, URIRef):
            _, group = group.rsplit("/", 1)
        elif isinstance(group, Literal):
            _, group = unicode(group)
        else:
            return False
        queue.add(dataset, "groups", [unicode(group)])
        return True
    return f



import sys
import urllib2
from urlparse import urlparse
from datetime import datetime
from logging import getLogger
from rdflib.Graph import ConjunctiveGraph, Graph
from rdflib import URIRef, Literal, Variable, BNode
from rdflib.Namespace import Namespace
from rdflib.store.SPARQL import SPARQLStore
from FuXi.Rete.Util import generateTokenSet
from curate.work import queue

RDF = Namespace("http://www.w3.org/1999/02/22-rdf-syntax-ns#")
RDFS = Namespace("http://www.w3.org/2000/01/rdf-schema#")
DCT = Namespace("http://purl.org/dc/terms/")
HTTP = Namespace("http://www.w3.org/2006/http#")
METH = Namespace("http://www.w3.org/2008/http-methods#")
CURL = Namespace("http://eris.okfn.org/ww/2010/12/curl#")

class Action(object):
    def __init__(self, sbind, obind):
        self.sbind = sbind
        self.obind = obind
        self.log = getLogger("%s(%s, %s)" % (self.__class__.__name__, sbind.n3(), obind.n3()))
        self.log.debug("init")
    def get(self, term, token):
        if isinstance(term, Variable):
            for binding in token.bindings:
                if term in binding:
                    return binding[term]
        return term

class httpReq(Action):
    def __call__(self, tNode, inferredTriple, token, _binding, debug = False):
        """
        Check if the given resource responds with data to an HTTP
        REQ request.
        """
        resource = self.get(self.sbind, token)
        method = self.get(self.obind, token)
        self.log.info("%s" % resource) 

        class _Request(urllib2.Request):
            def get_method(self):
                return method
        request = _Request(resource)

        #g = tNode.network.inferredFacts
        g = Graph()
        conn = BNode()
        g.add((conn, RDF["type"], HTTP["Connection"]))
        g.add((conn, DCT["date"], Literal(datetime.utcnow())))
        parsed = urlparse(resource)
        host, port = parsed.hostname, parsed.port
        if port is None:
            if parsed.scheme == "https:":
                port = 443
            else:
                port = 80
        g.add((conn, HTTP["connectionAuthority"], Literal("%s:%s" % (host, port))))

        req = BNode()
        g.add((conn, HTTP["requests"], req))
        g.add((req, RDF["type"], HTTP["Request"]))
        g.add((req, HTTP["methodName"], Literal("HEAD")))
        g.add((req, HTTP["mthd"], METH["HEAD"]))
        g.add((req, HTTP["requestURI"], resource))
        if request.headers:
            self.record_headers(g, req, request.headers)

        resp = BNode()
        g.add((req, HTTP["resp"], resp))
        g.add((resp, RDF["type"], HTTP["Response"]))
        
        try:
            response = urllib2.urlopen(request)
            g.add((resp, HTTP["statusCodeNumber"], Literal("%s" % response.getcode())))
            if response.headers:
                self.record_headers(g, resp, response.headers)
            content = response.read(4096)
            response.close()
        except urllib2.HTTPError, response:
            g.add((resp, HTTP["statusCodeNumber"], Literal("%s" % response.getcode())))
            if response.headers:
                self.record_headers(g, resp, response.headers)
            content = response.read(4096)
            response.close()
            self.log.error("%s %s" % (resource, response))
        except urllib2.URLError, e:
            g.add((resp, HTTP["statusCodeNumber"], Literal("-1")))
            g.add((conn, RDFS["comment"], Literal(e.reason)))
            self.log.error("%s %s" % (resource, e))

        tNode.network.feedFactsToAdd(generateTokenSet(g))
        tNode.network.inferredFacts += g

    def record_headers(self, g, msg, hdict):
        for h in hdict:
            head = BNode()
            g.add((msg, HTTP["headers"], head))
            g.add((head, RDF["type"], HTTP["MessageHeader"]))
            g.add((head, HTTP["fieldName"], Literal(h)))
            g.add((head, HTTP["fieldValue"], Literal(hdict[h])))

class curlReq(Action):
    def http_conn(self, g, resource, method):
        conn = BNode()
        g.add((conn, RDF["type"], HTTP["Connection"]))
        g.add((conn, DCT["date"], Literal(datetime.utcnow())))
        parsed = urlparse(resource)
        host = parsed.hostname
        try:
            port = parsed.port
        except:
            port = None
        if port is None:
            if parsed.scheme == "https:":
                port = 443
            else:
                port = 80
        g.add((conn, HTTP["connectionAuthority"], Literal("%s:%s" % (host, port))))

        req = BNode()
        g.add((conn, HTTP["requests"], req))
        g.add((req, RDF["type"], HTTP["Request"]))
        g.add((req, HTTP["methodName"], Literal(method)))
        g.add((req, HTTP["mthd"], METH[method]))
        g.add((req, HTTP["requestURI"], resource))
        resp = BNode()
        g.add((req, HTTP["resp"], resp))
        g.add((resp, RDF["type"], HTTP["Response"]))
        return resp

    def generic_conn(self, g, resource, method):
        conn = BNode()
        g.add((conn, RDF["type"], CURL["Curl"]))
        g.add((conn, DCT["date"], Literal(datetime.utcnow())))
        g.add((conn, CURL["uri"], resource))
        return conn

    def recode(self, s):
        try: return s.decode("utf-8")
        except: pass
        try: return s.decode("latin1")
        except: pass
        try: return s.decode("koi8-r")
        except: pass
        raise UnicodeError(s)

    def __call__(self, tNode, inferredTriple, token, _binding, debug = False):
        import pycurl
        resource = self.get(self.sbind, token)
        method = self.get(self.obind, token)

        self.log.info("Fetching %s" % resource)
        parsed_resource = urlparse(resource)

        g = Graph()


        c = pycurl.Curl()
        c.setopt(c.URL, str(resource))
        c.setopt(c.FAILONERROR, 0)
        c.setopt(c.FOLLOWLOCATION, 1)
        c.setopt(c.MAXREDIRS, 5)

        if parsed_resource.scheme in ("http", "https"):
            resp = self.http_conn(g, resource, method)
            def handle_header(h):
                h = h.strip()
                if ":" in h:
                    k, v = [x.strip() for x in h.split(":", 1)]
                    head = BNode()
                    g.add((resp, HTTP["headers"], head))
                    g.add((head, RDF["type"], HTTP["MessageHeader"]))
                    g.add((head, HTTP["fieldName"], Literal(k)))
                    g.add((head, HTTP["fieldValue"], Literal(v)))
            def handle_data(s):
                s = self.recode(s)
                g.add((resp, HTTP["body"], Literal(s)))
                return 0
            c.setopt(c.HEADERFUNCTION, handle_header)
            c.setopt(c.WRITEFUNCTION, handle_data)
            if method == u"HEAD":
                c.setopt(c.NOBODY, 1)
        else:
            resp = self.generic_conn(g, resource, method)
            def handle_data(s):
                s = self.recode(s)
                g.add((resp, CURL["data"], Literal(s)))
                return 0
            c.setopt(c.WRITEFUNCTION, handle_data)

        if parsed_resource.scheme in ("https"):
            c.setopt(c.SSL_VERIFYPEER, 0)

        success = False

        try:
            c.perform()
            success = True
        except pycurl.error, e:
            ## why do we have to do this manually here?
            sys.exc_clear()
            if e.args[0] == pycurl.E_WRITE_ERROR:
                ## we purposely fail the write in order not to
                ## download the whole thing
                success = True
            else:
                g.add((resp, RDFS["comment"], Literal(e.args[1])))

        if parsed_resource.scheme in ("http", "https"):
            code = c.getinfo(c.HTTP_CODE)
            g.add((resp, HTTP["statusCodeNumber"], Literal("%s" % code)))
            if code > 399:
                success = False
                
        if success:
            g.add((resp, CURL["status"], CURL["Success"]))
        else:
            g.add((resp, CURL["status"], CURL["Failure"]))            

        c.close()

        tNode.network.feedFactsToAdd(generateTokenSet(g))
        tNode.network.inferredFacts += g

class addTag(Action):
    def __call__(self, tNode, inferredTriple, token, _binding, debug = False):
        dataset = self.get(self.sbind, token)
        tag = self.get(self.obind, token)
        queue.add(dataset, "tags", [unicode(tag)])

class delTag(Action):
    def __call__(self, tNode, inferredTriple, token, _binding, debug = False):
        dataset = self.get(self.sbind, token)
        tag = self.get(self.obind, token)
        queue.remove(dataset, "tags", [unicode(tag)])

class addGroup(Action):
    def __call__(self, tNode, inferredTriple, token, _binding, debug = False):
        dataset = self.get(self.sbind, token)
        group = self.get(self.obind, token)

        if isinstance(group, URIRef):
            _, group = group.rsplit("/", 1)
        elif isinstance(group, Literal):
            _, group = unicode(group)
        else:
            return
        queue.add(dataset, "groups", [unicode(group)])




import pkg_resources
from rdflib.Graph import Graph
from rdflib import URIRef
from rdflib import Namespace
from FuXi.Rete.RuleStore import SetupRuleStore
from FuXi.Horn.HornRules import HornFromN3
from FuXi.Horn.PositiveConditions import Uniterm

CURATE = Namespace("http://eris.okfn.org/ww/2010/12/curate#")

import logging
log = logging.getLogger(__name__)

def getBuiltins():
    builtins = {}
    for ep in pkg_resources.iter_entry_points("curate.builtins"):
        log.info("Adding builtin curate:%s" % ep.name)
        builtin = ep.load()
        builtins[CURATE[ep.name]] = builtin
    return builtins

def getActions():
    actions = {}
    for ep in pkg_resources.iter_entry_points("curate.actions"):
        log.info("Adding action curate:%s" % ep.name)
        action = ep.load()
        actions[CURATE[ep.name]] = action
    return actions

def makeRuleStore(rules):
    builtins = getBuiltins()
    ruleStore, ruleGraph, network = SetupRuleStore(makeNetwork=True,
                                                   additionalBuiltins=builtins)

    for ruleFile in rules:
        log.info("Adding ruleset %s" % ruleFile)
        for rule in HornFromN3(ruleFile):
            network.buildNetworkFromClause(rule)

    actions = getActions()
    for tNode in network.terminalNodes:
        for rule in tNode.rules:
            if not isinstance(rule.formula.head, Uniterm):
                continue
            headTriple = rule.formula.head.toRDFTuple()
            for pred in actions:
                if headTriple[1] == pred:
                    action = actions[pred](headTriple[0], headTriple[2])
                    tNode.executeActions[headTriple] = (True, action)

    return ruleStore, ruleGraph, network



from ckanclient import CkanClient
log = __import__("logging").getLogger("work")

class _WorkQueue(object):
    def __init__(self):
        self.flush()

    def flush(self):
        self.updates = {}
        self.removals = {}

    def add(self, dataset, key, value):
        dsd = self.updates.setdefault(dataset, {})
        self.__merge__(dsd, { key: value })

    def remove(self, dataset, key, value):
        dsd = self.removals.setdefault(dataset, {})
        self.__merge__(dsd, { key: value })

    def __merge__(self, dst, src):
        for key, value in src.items():
            if isinstance(value, list):
                vl = dst.setdefault(key, [])
                [vl.append(x) for x in value if x not in vl]
            elif isinstance(value, dict):
                vd = dst.setdefault(key, {})
                vd.update(value)
            else:
                dst[key] = value

    def __unmerge__(self, dst, src):
        for key, value in src.items():
            if isinstance(value, list):
                vl = dst.setdefault(key, [])
                [vl.remove(x) for x in value if x in vl]
            elif isinstance(value, dict):
                vl = dst.setdefault(key, {})
                for k,v in value.items():
                    if vl.get(k) == v:
                        del vl[k]

    def __iter__(self):
        datasets = set()
        [datasets.add(x) for x in self.updates]
        [datasets.add(x) for x in self.removals]
        for k in datasets:
            yield k, self.updates.get(k, {})

    def __fixup__(self, pkg):
        balky = ("id", "metadata_created", "metadata_modified", "relationships",
                 "ratings_average", "ratings_count", "ckan_url")
        for remove in balky:
            if remove in pkg: del pkg[remove]
        
    def process(self, *av, **kw):
        ckan = CkanClient(*av, **kw)
        for dataset, descr in self:
            _, pkgname = dataset.rsplit("/", 1)
            pkg = ckan.package_entity_get(pkgname)
            self.__unmerge__(pkg, self.removals.get(dataset, {}))
            self.__merge__(pkg, descr)
            groups = pkg.get("groups", [])
            self.__fixup__(pkg)
            ckan.package_entity_put(pkg)
            self.log_api_result(pkgname, ckan)
            for groupname in groups:
                group = ckan.group_entity_get(groupname)
                pkglist = group.setdefault("packages", [])
                if pkgname not in pkglist:
                    pkglist.append(pkgname)
                ckan.group_entity_put(group)
                self.log_api_result(groupname, ckan)
        self.flush()

    def log_api_result(self, entity, ckan):
        if ckan.last_status != 200:
            message = " %s" % ckan.last_message
        else:
            message = ""
        log.info("PUT %s: %s%s" % (entity, ckan.last_status, message))

queue = _WorkQueue()



from setuptools import setup, find_packages
import sys, os

version = '0.11'
requirements=[
    "ckanclient",
    "pycurl",
    "rdflib",
    "FuXi>=1.2",
    ]

if isinstance(sys.version_info, tuple): # implies some values of 2.6
    requirements.append("argparse")
elif sys.version_info.major == 2 and sys.version_info.minor < 7:
    requirements.append("argparse")

setup(
    name='curate',
    version=version,
    description="Curation Utilities for CKAN",
    long_description="""\
Curation Utilities for CKAN""",
    # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 2.6",
        "Topic :: Internet",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords="ckan rdf rdflib",
    author='Open Knowledge Foundation',
    author_email='okfn-help@lists.okfn.org',
    url="http://packages.python.org/curate/",
    license='AGPL',
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    include_package_data=True,
    zip_safe=False,
    install_requires=requirements,
    entry_points="""
        # -*- Entry points: -*-
        [console_scripts]
        curate=curate.command:curate

        [curate.builtins]

        [curate.actions]
        httpReq=curate.actions:httpReq
        curlReq=curate.actions:curlReq
#        sparqlCheck=curate.builtins:sparqlCheck
        addTag=curate.actions:addTag
        delTag=curate.actions:delTag
        addGroup=curate.actions:addGroup
    """,
)



"""
.. program:: curate

Tool for curation of CKAN datasets. It reads RDF descriptions
of the dataset and applies the specified rules. Inferred 
statements are written to the standard output in Notation 3
and built-in predicates may be used to check specific conditions
or perform certain actions.

Usage: ``curate [options] [dataset [dataset [...]]]``

.. cmdoption:: -b base_uri

    Base URI to read dataset descriptions from. The RDF graph
    at ${base_uri}/${dataset} is read for each dataset and 
    passed as input to the inference rules.

.. cmdoption:: -a api_base

    Base URI for calls to the CKAN API. This is useful for
    running the tool against different CKAN instances.

.. cmdoption:: -r rules

    Inference rules in Notation 3 to read. May be specified
    multiple times

.. cmdoption:: -k api_key

    CKAN API Key to use where authentication is required

.. cmdoption:: -s

    Save any inferred dataset metadata to the CKAN instance.
    If this is not specified the program will run and perform all
    checks but will not actually perform any operations via the
    CKAN api.

.. cmdoption:: -v

    Verbose output

.. cmdoption:: -h, --help

    output a help message

"""

import argparse

import warnings
warnings.simplefilter("ignore", DeprecationWarning)
import logging

from curate.rules import makeRuleStore
from rdflib.Graph import Graph
from FuXi.Rete.Util import generateTokenSet
from ckanclient import CkanClient
from curate.work import queue

def curate():
    parser = argparse.ArgumentParser(description="""
Tool for curation of CKAN datasets. It reads RDF descriptions
of the dataset and applies the specified rules. Inferred 
statements are written to the standard output in Notation 3
and built-in predicates may be used to check specific conditions
or perform certain actions.
""")
    parser.add_argument("-b", dest="base", default="http://semantic.ckan.net/package/",
                        help="RDF description base URI to look for packages to consider")
    parser.add_argument("-r", dest="rules", action="append", default=[], 
                        help="N3 rules (can specify more than once)",)
    parser.add_argument("-k", dest="api_key", help="CKAN API Key")
    parser.add_argument("-a", dest="api_base", help="CKAN API base")
    parser.add_argument("-d", dest="delta", action="store_true", 
                        help="Accumulate closure delta")
    parser.add_argument("-v", dest="debug", action="store_true",
                        help="Verbose output")
    parser.add_argument("-s", dest="save", action="store_true",
                        help="Save inferred metadata back to CKAN")
    parser.add_argument("datasets", nargs="*", 
                        help="Dataset(s) to check")
    args = parser.parse_args()

    logging.basicConfig(
        level=logging.DEBUG if args.debug else logging.INFO,
        format="%(asctime)s [%(levelname)s] %(name)s: %(message)s"
    )
    log = logging.getLogger(__name__)

    ruleStore, ruleGraph, network = makeRuleStore(args.rules)

    datasets = args.datasets
    if not datasets:
        datasets = CkanClient(base_location=args.api_base, api_key=args.api_key
                              ).package_register_get()

    closureDelta = Graph()
    for dataset in datasets:
        log.info("processing %s" % dataset)
        network.reset(closureDelta)
        try:
            g = Graph()
            g.parse(args.base + dataset)
        except Exception, e:
            log.error("exception loading graph at %s" % (args.base+dataset,))
            continue
        network.feedFactsToAdd(generateTokenSet(g))

        if args.save:
            queue.process(base_location=args.api_base, api_key=args.api_key)

        if not args.delta:
            closureDelta = Graph()


    if args.delta:
        print closureDelta.serialize(format="n3")

